module GreedyKernelMethods

using DocStringExtensions
using KernelFunctions

# Types
export MaximizeDeterminant, 
       MaximizeResidual, 
       GreedyResidualNormMinimization
# Functions
export greedy_interpolation
export max_determinant

include("types.jl")

#######################################################################
#                           Main functions                            #
#######################################################################

"""
$(SIGNATURES)
Returns a set of `l` points chosen among the columns of `X` by greedily maximzing
the determinant of the associated kernel matrix for kernel `κ`.
"""
function max_determinant(κ, X, l)
    Z, _ = greedy_interpolation(MaximizeDeterminant(), κ, X, l; compute_weights=false)
    return Z    # Weights cannot be computed without f_X.
end

"""
greedy_interpolation(s::LandmarkSelectionStrategy, κ, X, l, f_X::T = nothing)
greedy_interpolation(s, κ, X, l, f::Function)

Builds a kernel interpolant of function `f` using `l` landmarks greedily chosen from `X` according to the strategy `s`.
Returns a tuple `(Z, w)` where `Z` are landmarks (d×l) and `w` a vector of weights (length l) or `nothing` depending on the value of `compute_weights`.

Arguments: 
- s: strategy to select landmarks.
- κ: kernel function.
- X: data (columns = samples).
- l: number of interpolation points.
- f_X: vector of evaluations [f(x₁),…,f(x_n)]ᵀ.

Keyword arguments:
- `compute_weights` (=`false`): boolean indicating whether computing the weights of f in the basis ϕ(z_1),…,ϕ(z_m).

Warning: Weights are computed iteratively but in a way which is not numerically stable. It is likely to work well for the first iterations but might perform poorly for larger values of `l`. For this reason, `compute_weights` defaults to `false`.
"""
greedy_interpolation

# Evaluate f if needed
greedy_interpolation(s, κ, X, l, f::Function) = greedy_interpolation(s,κ,X,l,[f(x) for x in eachcol(X)])
function greedy_interpolation(s::LandmarkSelectionStrategy, κ, X, l, f_X::T = nothing; 
                              compute_weights = false) where T<:Union{Vector,Nothing}
	n = size(X, 2)
	@assert l≤n "Cannot sample l=$l landmarks, only $(size(X,2)) samples provided."
    @assert (isnothing(f_X) || (length(f_X)==n)) "Dimensions of `X` and `f_X` do not match."
    @assert (~isnothing(f_X) || (s isa MaximizeDeterminant && ~compute_weights)) "f_X should be provided. (`nothing` is supported only for the MaximizeDeterminant strategy (which does not compute the residual) when compute_weights==false.)"

    # General notations:
    # - ϕ: canonical feature associated to κ
    # - Z=[z_1,…,z_l]=X[:,S]: landmarks, subset of columns of X 
    # - U=(u_j)_{1≤j≤l} Gram-schmidt basis of the (ϕ(z_j))_{1≤j≤l}
    #
    # Notations of the variables:
    # - C (l×n): C[:,i]=coefficients of ϕ(x_i) in the basis U constructed so far.
    #   Hence after update iteration k, C[1:k,i]=[u_j(x_i)]_{1≤j≤k} and C[k+1:end,:]=0.
    #   Filled row by row but rows are never updated after filling because U is an ONB.
    # - S (size n): Support (position/index in X of the points chosen as landmarks, Z=X[:,S])
    #   Thus it's a vector of integers in 1:n of size k at the end of iteration k.
    # - Pfun²_X (size n): squared power function evaluated at the (x_i)_{1≤i≤n}.
    #   i.e. Pfun²_X[i]=‖(I-Pm)ϕ(x_i)‖_H².
    #   At every iteration, det(K_{S∪{j}})=det(K_S)·Pfun²_X[i].
    # - r (size n): residual for the points X
    # - c_f (size l): coefficients of f
    # - B (l×l): Upper-triangular such that U = [ϕ(z_1),…,ϕ(z_m)] B. 
    #   Also Bᵀ=(C[:,S]ᵀ)^{-1}.
    
    ##### Initialization #####
	C = zeros(l,n)
    Pfun²_X = kernelmatrix_diag(κ, X; obsdim=2)     # Power function
    r = need_residual(s) ? copy(f_X) : nothing      # Residual
    c_f = need_residual(s) ? zeros(l) : nothing     # Coefficients of f in U. 
    B = compute_weights ? zeros(l,l) : nothing      
    @debug "greedy_interpolation (compute_weights=$(compute_weights), need_residual=$(need_residual(s)))"

    ##############################################
    S = [] # Support
    k = 0   # Iteration counter
	while k < l

        # @debug "Iteration $k/$l (BEGINING, max pfun²=$(maximum(Pfun²_X)))."
        ##### Select new atom − O(n) time #####
        j = try
            select_new_atom(s, Pfun²_X, r)
        catch e
            if isa(e, NoNumericallyStableNewAtomError)
                @warn "Asked for l=$l points but no numerically stable new atom can be selected from the data. Final solution has thus only l=$k points."
                break
            else
                rethrow(e)
            end
        end

        push!(S, j)
        k += 1
        # @debug "Iteration $k/$l (s=$s, j=$j, pfun²[j]=$(Pfun²_X[j]))."

        ##### Update the basis − O(n(d+l)) time #####
        # We could pre-compute everything but would be a waste if l<<n
		Kji = kernelmatrix(κ, X[:,j:j], X[:,:]; obsdim=2)   # 1×n (Θ(nd) if kernel = Θ(d))
        # ∀i∊1:n, we use <u_k,ϕ(x_i)> = ((I-P_{k-1})ϕ(x_i))(z_k) / Pfun_{k-1}(z_k)
        to_update = Pfun²_X .> eps()   # Update only points which are not already "almost" in the subspace (otherwise we get 0/0 on next line)
        C[k,to_update] .= (Kji[1,to_update] - vec(C[:,j]'*C[:,to_update])) / sqrt(Pfun²_X[j]) 	# 1×n (cost kn)

        ##### Update power function and residual − O(n) time #####
        if need_residual(s) 
            c_f[k] = r[j]/sqrt(Pfun²_X[j])
            r .-= c_f[k]*C[k,:]     # Update residual
        end
        Pfun²_X .-= vec(C[k,:].^2)
        Pfun²_X[j] = -Inf    # So that it will never be chosen again

        ##### Update B − O(l²) time (for all iterations total) #####
        if compute_weights
            B[1:k-1,k] = - (B[1:k-1,1:k-1] * C[1:k-1,j]) / C[k,j]   # O(k)=O(l)
            B[k,k] = 1 / C[k,j]     # O(1)
        end
	end
    w = compute_weights ? B[1:k,1:k]*(B[1:k,1:k]'*f_X[S]) : nothing  # B'f_X = c_f (but not always computed)
    @assert length(S) == k
    compute_weights && (@assert length(w) == k)
	return X[:,S], w
end

"""
select_new_atom(s::LandmarkSelectionStrategy, Pfun², r)

Returns the index of the newly chosen atom among n points according to strategy `s`, using the knowledge of the `r` and the squared power function `Pfun²` (both should be `n`-dimensional vectors).
"""
function select_new_atom(s, Pfun², r)
    stable_idx = (Pfun² .≥ eps())
    if sum(stable_idx) == 0
        throw(NoNumericallyStableNewAtomError())
    else
        c = new_atom_max_criterion(s, Pfun², r)
        potential_idx = (c .≈ maximum(c[stable_idx]))
        # @debug "Sampling new atom randomly among $(sum(potential_idx)) candidates."
        return rand(findall(potential_idx .& stable_idx))
    end
end
new_atom_max_criterion(s::MaximizeDeterminant, Pfun², r) = Pfun²
new_atom_max_criterion(s::MaximizeResidual, Pfun², r) = r
new_atom_max_criterion(s::GreedyResidualNormMinimization, Pfun², r) = r.^2 ./ Pfun²


need_residual(s) = true
need_residual(::MaximizeDeterminant) = false

end # module 
